unit SetUtils;
{$I Common.inc}
//************************************************************************
//                                           *
//                                                                       *
//************************************************************************

interface

uses Root, Items, SetList;

//   

function GenerateChars(arg: integer): TSet;

//   

function GenerateNumbers(arg: integer): TSet;

//         
// aBuf  -  

function GenCostItems(aCost: integer; aBuf: TBuffer): TBuffer;

///       

function CalcCost(aSet: TSet): integer;

//      

procedure RemoveItemsFromBuf(aBuf: TBuffer; aSet: TSet);

//     

function GenUniversum(aSet: TSet): TSet;

//    

function CalcUniversumPower(aSet: TSet): integer;

//   ()   

function ReadSetsFromText(const aName: string; BaseSet: TSet): TBuffer;

//////////////////////////////////////////////////////////////////////////
implementation

//////////////////////////////////////////////////////////////////////////
//   
//////////////////////////////////////////////////////////////////////////

function GenerateChars(arg: integer): TSet;
var i: integer;
    c: char;
begin
  Result:= CreateSet;
  c:='a';
  for i:= 1 to arg do begin
    Result.Insert(TItemChar.Create(c));
    c:= Succ(c);
  end;
end;

//////////////////////////////////////////////////////////////////////////
//   
//////////////////////////////////////////////////////////////////////////

function GenerateNumbers(arg: integer): TSet;
var i: integer;
begin
  Result:= CreateSet;
  for i:= 1 to arg do
    Result.Insert(TItemNum.Create(i));
end;

//////////////////////////////////////////////////////////////////////////
//         
// aBuf  -  
// aCost -  :
//   0 -      1  99
//   1 -    
//   2 -  ,   
//   3 -    
//   4 -     
//////////////////////////////////////////////////////////////////////////

function GenCostItems(aCost: integer; aBuf: TBuffer): TBuffer;
var S: TSet;
    cost: integer;
    n: integer;
begin
  Result:= TBuffer.Create;
  S:= aBuf.Get as TSet;
  while Assigned(S) do begin
    n:= S.GetCount;
    case aCost of
     1: cost:= 1;
     2: cost:= 1 + (n div 2) + Random(n);
     3: cost:= n;
     4: cost:= 1 + (n-1)*(n-1);
     else cost:= 1 + Random(99);
    end;
    Result.Put(TCostSet.Create(cost, S, true));
    S:= aBuf.Get as TSet;
  end;
end;

//////////////////////////////////////////////////////////////////////////
///       
//////////////////////////////////////////////////////////////////////////

function CalcCost(aSet: TSet): integer;
var t : TCostSet;
begin
  Result:= 0;
  aSet.PositionPush;
  t:= aSet.GetFirst as TCostSet;
  while Assigned(t) do begin
    if t.InheritsFrom(TCostSet)
      then Inc(Result, t.mCost);
    t:= aSet.GetNext as TCostSet;
  end;
  aSet.PositionPop;
end;

//////////////////////////////////////////////////////////////////////////
//     
//////////////////////////////////////////////////////////////////////////

function GenUniversum(aSet: TSet): TSet;
var t : TItem;
    CS: TCostSet;  //   
begin
  Result:= CreateSet;
  aSet.PositionPush;
  CS:= TCostSet(aSet.GetFirst);
  while Assigned(CS) do begin
    t:= CS.mSet.GetFirst;
    while Assigned(t) do begin
      Result.Insert(t);
      t:= CS.mSet.GetNext;
    end;
    CS:= TCostSet(aSet.GetNext);
  end;
  aSet.PositionPop;
end;

//////////////////////////////////////////////////////////////////////////
//     
//////////////////////////////////////////////////////////////////////////

function CalcUniversumPower(aSet: TSet): integer;
var S: TSet;
begin
  S:= GenUniversum(aSet);    //  -
  Result:= S.GetCount;       //    
  S.Free;
end;

//////////////////////////////////////////////////////////////////////////
//     aBuf   aSet
//////////////////////////////////////////////////////////////////////////

procedure RemoveItemsFromBuf(aBuf: TBuffer; aSet: TSet);
var   CS: TItem;  //     (CostSet)
      S : TSet;   //   aSet
      t : TItem;  //   aSet

begin
  if not Assigned(aBuf) or not Assigned(aSet) then Exit;
  //   aSet
  S:= CreateSet;
  S.CopyItems(aSet);  // S := aSet
  //     aSet
  while S.GetCount > 0 do begin
    CS:= aBuf.Get;                    //   
    if not Assigned(CS) then Break;   // ,   
    //  CS   S
    t:= S.GetFirst;
    while Assigned(t) do begin
       if t = CS then break;     // 
       t:= S.GetNext;
    end;
    if t = CS
      then S.Delete(CS)   //  ,     aSet
      else aBuf.Put(CS);  //     
  end;
  S.Free;
end;

//////////////////////////////////////////////////////////////////////////
//   ()   
//////////////////////////////////////////////////////////////////////////

function ReadSetsFromText(const aName: string; BaseSet: TSet): TBuffer;
var F : TextFile;
    S : string;
    CS: TCostSet;

    function MakeSet: TCostSet;
    var i, cost: integer;
        ChrItem : TItemChar;
        Item    : TItemChar;
        ChrSet  : TSet;

    begin
      Result:= nil;
      if (Pos('{',S) = 0) or (Pos('}',S) = 0) then Exit;
      i:=1;
      while S[i] = #32 do Inc(i);
      cost:=0;
      while S[i] in ['0'..'9'] do begin
        cost:= 10*cost + Ord(S[i]) - Ord('0');
        Inc(i);
      end;
      ChrSet:= CreateSet;
      while i <= Length(S) do begin
        if S[i] in ['a'..'z'] then begin
          ChrItem:= TItemChar.Create(S[i]);
          Item:= TItemChar(BaseSet.GetObject(ChrItem));
          if Assigned(Item) then begin //     
            ChrItem.Free;              //   
            ChrSet.Insert(Item);       //   
          end else begin               //       
            ChrSet.Insert(ChrItem);    //   
            BaseSet.Insert(ChrItem);   //     
          end;
        end;
        Inc(i);
      end;
      Result:= TCostSet.Create(cost, ChrSet, true);
    end;

begin  { ReadSetsFromText }

  Result:= TBuffer.Create;
  AssignFile(F, aName); Reset(F);
  while not Eof(F) do begin
    Readln(F,S);
    CS:= MakeSet;
    if Assigned(CS) then Result.Put(CS);
  end;
  Close(F);
end;

//////////////////////////////////////////////////////////////////////////

end.
